using System;
using System.Collections.Generic;
using System.Windows.Forms;

using System.Runtime.InteropServices;
using System.Drawing;
using System.Drawing.Printing;

namespace printers
{
    static class Program
    {
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main()
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            PrinterHelper.DisplayPrinters();
        }
    }

    public class PrinterHelper
    {
        /// <summary>
        /// Metoda koja vraca ime porta na koji je fizicki prikljucen
        /// </summary>
        /// <param name="printerSettings"></param>
        /// <returns></returns>
        public static String GetPrinterSettingsOutputPort(PrinterSettings printerSettings)
        {
            String outputPort = "";
            /*
            try {
                System.Reflection.FieldInfo fi = printerSettings.GetType().GetField("outputPort", System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic);
                outputPort = (String)fi.GetValue(printerSettings);
            } catch(Exception) { }
            */
            IntPtr pHandle = IntPtr.Zero;
            OpenPrinter(printerSettings.PrinterName, out pHandle, 0);
            uint cbNeeded = 0;
            bool bRet = GetPrinter(pHandle, 2, IntPtr.Zero, 0, ref cbNeeded);
            if(cbNeeded > 0) {
                IntPtr pAddr = Marshal.AllocHGlobal((int)cbNeeded);
                bRet = GetPrinter(pHandle, 2, pAddr, cbNeeded, ref cbNeeded);
                if(bRet) {
                    PRINTER_INFO_2 Info2 = new PRINTER_INFO_2();
                    Info2 = (PRINTER_INFO_2)Marshal.PtrToStructure(pAddr, typeof(PRINTER_INFO_2));
                    outputPort = Info2.pPortName;
                }
                Marshal.FreeHGlobal(pAddr);
            }
            ClosePrinter(pHandle);
            return outputPort;
        }


        /// <summary>
        /// metoda koja vraca fizicke margine koje stampac moze da odstampa
        /// </summary>
        /// <param name="hdc"></param>
        /// <param name="capindex"></param>
        /// <returns></returns>
        public static Margins GetHardMargins(PrinterSettings ps, bool isLandscape)
        {
            Margins m;
            Graphics e = ps.CreateMeasurementGraphics();
            float dpiX = e.DpiX;
            float dpiY = e.DpiY;

            IntPtr hDC = e.GetHdc();
            int offsetx = GetDeviceCaps(hDC, DeviceCap.PHYSICALOFFSETX);
            int offsety = GetDeviceCaps(hDC, DeviceCap.PHYSICALOFFSETY);
            int horzres = GetDeviceCaps(hDC, DeviceCap.HORZRES);
            int vertres = GetDeviceCaps(hDC, DeviceCap.VERTRES);
            int physicalwidth = GetDeviceCaps(hDC, DeviceCap.PHYSICALWIDTH);
            int physicalheight = GetDeviceCaps(hDC, DeviceCap.PHYSICALHEIGHT);
            e.ReleaseHdc(hDC);
            e.Dispose();

            int left = (int)Math.Ceiling((float)offsetx * 100.0F / dpiX);
            int top = (int)Math.Ceiling((float)offsety * 100.0F / dpiY);
            int right = (int)Math.Ceiling((float)(physicalwidth - horzres - offsetx) * 100.0F / dpiY);
            int bottom = (int)Math.Ceiling((float)(physicalheight - vertres - offsety) * 100.0F / dpiY);

            if(isLandscape && ps.LandscapeAngle == 90) {
                m = new System.Drawing.Printing.Margins(bottom, top, left, right);
            } else if(isLandscape && ps.LandscapeAngle == 270) {
                m = new System.Drawing.Printing.Margins(top, bottom, right, left);
            } else {
                m = new System.Drawing.Printing.Margins(left, right, top, bottom);
            }
            return m;
        }

        public static Boolean IsMatrixPrinter(PrinterSettings printerSettings)
        {
            Graphics e = printerSettings.CreateMeasurementGraphics();
            IntPtr hdc = e.GetHdc();
            int numColors = GetDeviceCaps(hdc, PrinterHelper.DeviceCap.NUMCOLORS);
            int trueType = PrinterHelper.DeviceCapabilities(printerSettings.PrinterName, "", PrinterHelper.DeviceCapabilitiesFlags.DC_TRUETYPE, IntPtr.Zero, IntPtr.Zero);
            e.ReleaseHdc(hdc);
            e.Dispose();
            return (numColors <= 16 && trueType == DCTT_BITMAP);
        }

        public static void DisplayPrinters()
        {
            PrinterSettings ps = new PrinterSettings();
            String S = "Total printers: " + PrinterSettings.InstalledPrinters.Count.ToString() + "\r\n\r\n";
            foreach(String printer in PrinterSettings.InstalledPrinters) {
                ps.PrinterName = printer;
                S += "Printer: " + ps.PrinterName + "\r\n";
                S += "port: " + PrinterHelper.GetPrinterSettingsOutputPort(ps) + "\r\n";

                Graphics e = ps.CreateMeasurementGraphics();
                IntPtr hdc = e.GetHdc();
                S += "margins: " + PrinterHelper.GetHardMargins(ps, false).ToString() + "\r\n";
                S += "Colors: " + PrinterHelper.GetDeviceCaps(hdc, PrinterHelper.DeviceCap.NUMCOLORS).ToString() + "\r\n";
                S += "TRUETYPE: " + PrinterHelper.DeviceCapabilities(ps.PrinterName, "", PrinterHelper.DeviceCapabilitiesFlags.DC_TRUETYPE, IntPtr.Zero, IntPtr.Zero).ToString() + "\r\n";
                S += "IsMatrix: " + PrinterHelper.IsMatrixPrinter(ps).ToString() + "\r\n\r\n";
                e.ReleaseHdc(hdc);
                e.Dispose();
            }

            System.Windows.Forms.Form form = new System.Windows.Forms.Form();
            System.Windows.Forms.TextBox textBox = new System.Windows.Forms.TextBox();
            textBox.Dock = DockStyle.Fill;
            textBox.Multiline = true;
            textBox.ScrollBars = ScrollBars.Both;
            textBox.Text = S;
            form.Text = "Installed Printers";
            form.Controls.Add(textBox);
            form.Size = new Size(300, 500);
            form.StartPosition = FormStartPosition.CenterScreen;
            form.ShowDialog();
            form.Dispose();
        }


        [DllImport("winspool.drv", SetLastError = true)]
        public static extern bool GetPrinter(IntPtr hPrinter, Int32 dwLevel, IntPtr pPrinter, Int32 dwBuf, out Int32 dwNeeded);

        [DllImport("winspool.drv", SetLastError = true)]
        public static extern bool GetPrinter(IntPtr hPrinter, uint dwLevel, IntPtr pPrinter, uint dwBuf, ref uint dwNeeded);

        [DllImport("winspool.Drv", EntryPoint = "OpenPrinterA", SetLastError = true, CharSet = CharSet.Ansi, ExactSpelling = true, CallingConvention = CallingConvention.StdCall)]
        public static extern bool OpenPrinter([MarshalAs(UnmanagedType.LPStr)] string szPrinter, out IntPtr hPrinter, Int32 pDefault);

        [DllImport("winspool.drv", SetLastError = true)]
        public static extern int ClosePrinter(IntPtr hPrinter);

        [DllImport("gdi32.dll")]
        public static extern int GetDeviceCaps(IntPtr hdc, DeviceCap cap);

        [DllImport("winspool.drv", SetLastError = true)]
        public static extern Int32 DeviceCapabilities(string device, string port, DeviceCapabilitiesFlags capability, IntPtr outputBuffer, IntPtr deviceMode);

        [StructLayout(LayoutKind.Sequential)]
        public struct PRINTER_INFO_2
        {
            public string pServerName;
            public string pPrinterName;
            public string pShareName;
            public string pPortName;
            public string pDriverName;
            public string pComment;
            public string pLocation;
            public IntPtr pDevMode;
            public string pSepFile;
            public string pPrintProcessor;
            public string pDatatype;
            public string pParameters;
            public IntPtr pSecurityDescriptor;
            public uint Attributes;
            public uint Priority;
            public uint DefaultPriority;
            public uint StartTime;
            public uint UntilTime;
            public uint Status;
            public uint cJobs;
            public uint AveragePPM;
        }

        public enum DeviceCap
        {
            DRIVERVERSION = 0,
            TECHNOLOGY = 2,
            HORZSIZE = 4,
            VERTSIZE = 6,
            HORZRES = 8,
            VERTRES = 10,
            BITSPIXEL = 12,
            PLANES = 14,
            NUMBRUSHES = 16,
            NUMPENS = 18,
            NUMMARKERS = 20,
            NUMFONTS = 22,
            NUMCOLORS = 24,
            PDEVICESIZE = 26,
            CURVECAPS = 28,
            LINECAPS = 30,
            POLYGONALCAPS = 32,
            TEXTCAPS = 34,
            CLIPCAPS = 36,
            RASTERCAPS = 38,
            ASPECTX = 40,
            ASPECTY = 42,
            ASPECTXY = 44,

            LOGPIXELSX = 88,
            LOGPIXELSY = 90,

            SIZEPALETTE = 104,
            NUMRESERVED = 106,
            COLORRES = 108,

            // Printing related DeviceCaps. These replace the appropriate Escapes
            PHYSICALWIDTH = 110,
            PHYSICALHEIGHT = 111,
            PHYSICALOFFSETX = 112,
            PHYSICALOFFSETY = 113,
            SCALINGFACTORX = 114,
            SCALINGFACTORY = 115
        }

        public enum DeviceCapabilitiesFlags : short
        {

            DC_FIELDS = 1,
            DC_PAPERS = 2,
            DC_PAPERSIZE = 3,
            DC_MINEXTENT = 4,
            DC_MAXEXTENT = 5,
            DC_BINS = 6,
            DC_DUPLEX = 7,
            DC_SIZE = 8,
            DC_EXTRA = 9,
            DC_VERSION = 10,
            DC_DRIVER = 11,
            DC_BINNAMES = 12,
            DC_ENUMRESOLUTIONS = 13,
            DC_FILEDEPENDENCIES = 14,
            DC_TRUETYPE = 15,
            DC_PAPERNAMES = 16,
            DC_ORIENTATION = 17,
            DC_COPIES = 18,
            DC_BINADJUST = 19,
            DC_EMF_COMPLIANT = 20,
            DC_DATATYPE_PRODUCED = 21,
            DC_COLLATE = 22,
            DC_MANUFACTURER = 23,
            DC_MODEL = 24,
            DC_PERSONALITY = 25,
            DC_PRINTRATE = 26,
            DC_PRINTRATEUNIT = 27,
            DC_PRINTERMEM = 28,
            DC_MEDIAREADY = 29,
            DC_STAPLE = 30,
            DC_PRINTRATEPPM = 31,
            DC_COLORDEVICE = 32,
            DC_NUP = 33
        }

        public const int DCTT_BITMAP = 1;
    }

}